/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include "ScatterWeb.System.h"

volatile UINT8 serial_buf1[SERIAL_BUFSIZE]; // The serial ISR writes bytes into this buffer.
volatile UINT8 serial_buf2[SERIAL_BUFSIZE];
volatile UINT8* serial_line;
volatile UINT8* buf;
volatile UINT8* rts_port;
volatile UINT8* cts_port;
volatile int serial_pos; // Current position in receive buffer ::serial_buffer.

#define SERIAL_STOP       do{ *cts_port |=  cts_bit; }while(0)  // Set CTS to 1 to stop receiving bytes.
#define SERIAL_RESUME     do{ *cts_port &= ~cts_bit; }while(0)  // Set CTS to 0 to resume receiving bytes.

UINT16 ta1;
UINT8  rts_bit;
UINT8  cts_bit;
UINT8  baudrate_U1;
UINT8 Comm_logLevel;

void _Comm_sendRS232(UINT8 data) { // sends one byte
  while (!(IFG2 & UTXIFG1)); // loop until USART1 TX buffer ready
  TXBUF1 = data;
}

bool Comm_print(const char *fmt, ...) {
  if((Comm_logLevel & 0x0F) >= NO) {
	va_list argp;
    va_start(argp, fmt);
    String_abstractWrite(_Comm_sendRS232, fmt, argp);
    va_end(argp);
  }
  return True;
}

bool Comm_log(UINT8 level, const char* fmt, ...) {
  va_list argp;
  if((Comm_logLevel & 0x0F) >= level) {
    va_start(argp, fmt);
    String_abstractWrite(_Comm_sendRS232, fmt, argp);
    va_end(argp);
  }
  return True;
}

bool Comm_clear() {
  runModule &= ~MF_SERIAL;
  serial_line = 0;
  SERIAL_RESUME;
  return True;
}

// Serial ISR-Routine for receiving bytes.
interrupt(NOVECTOR) serial_rec() {
  UINT8 temp = RXBUF1;
  if(temp=='\r') return;    // filter out carriage returns
  if(temp=='\n') {          // line is complete
    SERIAL_STOP;            // set CTS to 1 to stop receiving bytes
    buf[serial_pos] = '\0'; // terminate
    if(buf==serial_buf1) {
      buf         = serial_buf2;
      serial_line = serial_buf1;
    }
    else{
      buf         = serial_buf1;
      serial_line = serial_buf2;
    }
    WAKEUPLPM1(MF_SERIAL);
    serial_pos = 0;
  }
  else {
    if(temp=='\b') {
      if(serial_pos>0) serial_pos--; // backspace
    }
    else { // write byte into buf
      if(serial_pos==SERIAL_BUFSIZE) return;   // don't write anymore if buffer is full
      buf[serial_pos++]=temp;
    }
  }
}

interrupt(NOVECTOR) autorxIRQ() {
  UINT8 temp = RXBUF1;
  IE2 &= ~URXIE1; // Disable USART1 RX interrupt
  P1OUT &= ~0x10;
  switch (temp) {
    case 0x0D :
      baudrate_U1 = BPS_115200;
      Comm_init(baudrate_U1);
      break; // 115200
    case 0xE6 :
      baudrate_U1 = BPS_57600;
      Comm_init(baudrate_U1);
      break; // 57600
    case 0x1C :
      baudrate_U1 = BPS_38400;
      Comm_init(baudrate_U1);
      break; // 38400
    case 0xE0 :
      baudrate_U1 = BPS_19200;
      Comm_init(baudrate_U1);
      break; // 19200
    case 0x80 :
      baudrate_U1 = BPS_14400;
      Comm_init(baudrate_U1);
      break; // 14400
    default :
      baudrate_U1 = BPS_115200;
      Comm_init(baudrate_U1);
      break;
  }
  RXBUF1 = 0;    // clear RX Buffer from Dummy-Inputs
  IE2 |= URXIE1; // Enable USART1 RX interrupt
}

interrupt(NOVECTOR) binrxIRQ() {
}

bool Comm_init(UINT8 bps) {
  UINT16 i;
  //////////// VARS INIT ////////////////////////////////////////////////////
  Comm_logLevel = Configuration.boot_logLevel;
  Comm_clear();
  // init vars
  serial_pos = 0;
  for(i=0; i<SERIAL_BUFSIZE; i++) {
    serial_buf1[i] = 0x00;
    serial_buf2[i] = 0x00;
  }
  serial_line = 0;
  buf = serial_buf1;
  // Standard is to not use RTS/CTS by setting rts_bit and cts_bit to zero.
  rts_port = &P1IN;
  cts_port = &P1OUT;
  rts_bit = 0;
  cts_bit = 0;
  P3SEL |= 0xC0; // No select.
  //////////////////////////////////////// RS232
  UCTL1  = SWRST;
  UCTL1 |= CHAR; // 8-bit character
  switch(bps) {
    case BPS_2400:
      UTCTL1 = SSEL1;
      UBR01 = 0x55;
      UBR11 = 0x07;
      UMCTL1 = 0x09;
      break;
      // RS232 with 2400 bps @ 32768 Hz ACLK
    case BPS_4800:
      U0TCTL = SSEL1;
      UBR01 = 0xAA;
      UBR11 = 0x03;
      UMCTL1 = 0x6D;
      break; // 4800
    case BPS_9600:
      UTCTL1 = SSEL1;
      UBR01 = 0xD5;
      UBR11 = 0x01;
      UMCTL1 = 0x52;
      break; // uart1 4505600Hz 14399bps
    case BPS_14400:
      UTCTL1 = SSEL1;
      UBR01 = 0x38;
      UBR11 = 0x01;
      UMCTL1 = 0xEF;
      break; // uart1 4505600Hz 14399bps
    case BPS_19200:
      UTCTL1 = SSEL1;
      UBR01 = 0xEA;
      UBR11 = 0x00;
      UMCTL1 = 0x5B;
      break; // uart1 4505600Hz 19197bps - RS232 with 19200 bps @  Mhz
    case BPS_38400:
      UTCTL1 = SSEL1;
      UBR01 = 0x75;
      UBR11 = 0x00;
      UMCTL1 = 0x52;
      break; // uart1 4505600Hz 38378bps - RS232 with 38400 bps @  Mhz
    case BPS_57600:
      UTCTL1 = SSEL1;
      UBR01 = 0x4E;
      UBR11 = 0x00;
      UMCTL1 = 0x44;
      break; // uart1 4505600Hz 57616bps - RS232 with 57600 bps @  Mhz
    case BPS_115200:
      UTCTL1 = SSEL1;
      UBR01 = 0x27;
      UBR11 = 0x00;
      UMCTL1 = 0x10;
      break; // uart1 4505600Hz 115232bps - RS232 with 115200 bps @  Mhz
  }
  ME2 |= UTXE1 + URXE1; // Enable USART1 TXD/RXD, disabling does not yield any powersavings
  UCTL1  &= ~SWRST;     // Last thing is reset
  SET_ISR(UART1RX_VECTOR, serial_rec);
  IE2 |= URXIE1;        // Enable USART1 RX interrupt  
  return True;
}

COMMAND(brt, 0) {
  Comm_init(String_parseInt(&str[4], NULL));
}
