/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

/** \file ScatterWeb.Messaging.h
 ** \ingroup System
 ** \brief The ScatterWeb.Messaging provides fuction that allow you to connect
 ** to, monitor, and message on the network and send, receive, or peek messages. 
 ** <p>
 ** Control the MSP via RS232 or radio with simple ascii commands.
 ** <p>
 ** Remote command execution
 ** To issue an command over radio write @[id], answers will get [id] prepended.
 ** @85 rid
 ** [85] 85
 ** @85 tim
 ** [85] 23080608,984 2001-09-25 03:16:48
 ** <p>
 ** Command creation
 ** Use the COMMAND macro to easily create a new terminal command. The first
 ** argument to the Macros is always the three digit terminal command code. The
 ** second argument to the Macros is a 8bit flag field. For example the
 ** following:
 ** COMMAND(xxx, 0){ print("test\r\n"); }
 ** <p>
 ** will evaluate to:
 ** void term_xxx (const UINT8* str);
 ** __attribute__((section(".commands"))) const command_t com_xxx = { term_xxx, "xxx", 0 }; \
 ** void term_xxx (const UINT8* str) { print("test\r\n"); }
 ** <p>
 ** Besides creating the function header it first creates a forward declaration
 ** which is then needed in the following definition of a const command_t
 ** com_xxx struct in the .commands segment. com_xxx contains a pointer to the
 ** function, the three digit function code and a flag field. 
 ** Because all those command_t entries are put into a special segment
 ** (.commands), the function doCommand can easily loop through this segment
 ** (going from __term_start to __term_end also provied by the linker script)
 ** and search for the command.
 ** This even makes it possible, to declare terminal commands where needed, not
 ** only in terminal.c! (The firmware commands will still stay centralized in
 ** ScatterWeb.Messaging.c to have code & documentation of the firmware commands
 ** in one place.)
 ** <p>
 ** The needed modifications of the linker script can be seen in
 ** System/ldscript.x:
 ** \dontinclude ldscript.x
 ** \skip .text :
 ** \until   }  > text
 ** <p>
 ** All commands must write their answer to term_reply and it should not be
 ** longer than TERM_BUFSIZE. If nothing is written to term_reply automatically
 ** an "OK" is generated.
 ** <p>
 ** The only exception to that is when the bit SERIALONLY in the flag field is
 ** set. That means that the terminal command uses the serial functions
 ** directly (not writing its answer to term_reply) and can't be called over
 ** radio.
 ** 
 ** @author        Marco Hutta
 ** @version       1.0
 ** @since         1.0    
 **/

#ifndef ScatterWeb_MESSAGING
#define ScatterWeb_MESSAGING

/**
 * Command with this flag access uart directly with prints.
 *
 * @since         1.0
 */
#define SERIALONLY    (0x02)

/**
 * Use this MACRO to declare terminal command functions
 *
 * @since         1.0
 */
#define COMMAND(x, flags)   void term_##x (const UINT8* str); \
                            __attribute__((section(".commands"))) const command_t com_##x = { term_##x, #x, flags }; \
                            void term_##x (const UINT8* str)

/**
 * Terminal command function pointer type, argument is the command.
 *
 * @since         1.0
 */
typedef void(*fp_term)(const UINT8*);

/**
 * Reply buffer. All commands (except SERIALONLY) build their reply here.
 *
 * @since         1.0
 */
extern UINT8 term_reply[TERM_BUFSIZE];

/**
 * Struct for the command directory
 * <ul>
 * <li>function
 * <li>code[3]
 * <li>flags
 * </ul>
 *  
 * @since         1.0
 */  
typedef struct { 
  fp_term function;
  UINT8 code[3];
  UINT8 flags;
} command_t;

/**
 * Enables the Messaging
 *
 * @return        <code>True</code> always.
 * @since         1.0
 */
void Messaging_on();

/**
 * Filters packets for Messaging
 *
 * @return        <code>True</code> always. 
 * @since         1.0
 */
bool Messaging_radioHandler();

#endif
