/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include "ScatterWeb.System.h"

// 256 Ringbuffer macro implemenation.
#define RB(name)   \
    volatile struct {     \
      UINT16 guard1;      \
      UINT8 buffer[256];  \
      UINT8 head, tail;   \
      UINT16 guard2;      \
    } name

// Gets the first byte of the buffer and remove it.
#define RB_GET(b, x) do { x = (b).buffer[(b).head]; (b).head++; } while(0)

// Get current length.
#define RB_LENGTH(b) ((UINT8)( (b).tail - (b).head ))

// Appends another bytes to the buffer.
#define RB_PUT(b, x) do { (b).buffer[(b).tail] = x; (b).tail++; } while(0)

// Initializes the ringbuffer.
#define RB_INIT(b) do { (b).head = 0; (b).tail = 0; (b).guard1 = 0xABCD; (b).guard2 = 0xABCD; } while(0)

// TX
RB(txPackets);                    // Interrupt byte radio ringbuffer.
volatile UINT8 numPackets;        // Current number of packets in radio tx buffer.
volatile UINT8 txNum;             // Current tx Number.
volatile UINT16 txBufPos;         // Current position in txBuf.
unsigned char txLockStatus;       // Saves last value from SetupCC1020TX.
volatile UINT8 txBuf[140];        // Buffer for bulding packets with preamble.
volatile UINT8 ackBuf[18];        //   ""          ""      ACKS
volatile UINT16 txBufLength;      // How much to send in TxBuffer
volatile UINT16 txCallback;       // Sotres callback what to do after a packet has been sent.
volatile UINT16 txCallbackValue;  // Stores values to callback.
volatile UINT16 txBackoff;        // Var for storing backoff in states TX_WAITING, TX_BACKOFF.
volatile UINT16 timesRepeated;    // Var for storing how often a apcket hast been repeated during sending.
volatile UINT16 ackTimeout;       // Timeout how long to wait for an ack in RADIO_STROKE ticks, depends on baudrate and should come with the config.
volatile UINT8 txMode;
volatile UINT8 radioEnabled;
UINT8 isSending;                  // Show that tx state machine is sending!
UINT16 v0, v1, v2;  
UINT16 tim1, tim2, tim3, tim4;

#define TIMESTART()   do { tim1=TAR; tim2=TBR; }while(0)
#define TIMESTOP()    do { tim3=TAR; tim4=TBR; print("TIM %i %i\n", tim3-tim1, tim4-tim2); }while(0)
#define MAX_REPEATED  (5)         // How often a packet get repeated.
#define RADIO_STROKE  (9011)      // ~ 2 ms @ 4505600 Hz.

// RX
UINT8* rxBuf;
UINT8 rxBuf1[140];              // Buffer for received packets. FF=Empty 00=Rec 01=FULL
UINT8 rxBuf2[140];              // Buffer for received packets.
UINT8 rxBufPos;                 // Position in this buffer.
volatile time_t rxTime;         // When Synch byte was discovered.
volatile UINT8 rssi;            // Received Signal Strength Indicator read from chipcon during receive.
volatile UINT16 checksumR;      // Received checksum.
volatile UINT16 checksumC;      // Computed checksum.

// CRC
#define INIT_FCS  0xffff    // Initial FCS value, use this as starting value for fcs. in ::crc_calc.

/*
 *   brief CRC generating lookup table.
 *   This mysterious table is just the CRC of each possible byte.  It can be
 *   computed using the standard bit-at-a-time methods.  The polynomial can
 *   be seen in entry 128, 0x8408.  This corresponds to x^0 + x^5 + x^12.
 *   Add the implicit x^16, and you have the standard CRC-CCITT.
 */
UINT16 const ccitt_crc16_table[256] = {
 	0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
 	0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
 	0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
 	0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
 	0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
 	0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
 	0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
 	0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
 	0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
 	0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
 	0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
 	0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
 	0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
 	0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
 	0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
 	0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
 	0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
 	0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
 	0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
 	0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
 	0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
 	0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
 	0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
 	0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
 	0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
 	0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
 	0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
 	0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
 	0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
 	0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
 	0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
 	0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

// Recompute the FCS with one more character appended.
UINT16 _Net_ccittFcs(UINT16 fcs, UINT8 c) {
  return (((fcs) >> 8) ^ ccitt_crc16_table[((fcs) ^ (c)) & 0xff]);
}

// Forward declarations
void TX_STOPPED();
void TX_IDLE();
void TX_DONE();
void rxsearching();
void rxsynched0();
void rxsynched1();
void rxsynched2();
void rxsynched3();
void rxsynched4();
void rxsynched5();
void rxsynched6();
void rxsynched7();
void rxreceived();
void acktxdma();
void TX_BACKOFF();
void TX_WAITING();
void TX_WAITACK();

volatile UINT16 irqLock;

#define DISABLE_RX_IRQ() do { IE1 &= ~(URXIE0); } while(0)
#define ENABLE_RX_IRQ() do { IFG1 &= ~URXIFG0; IE1 |= URXIE0; } while(0)
#define DIRQSTART()  
#define DIRQEND()    

#define CC_LOCK (P2IN&0x04)
#define LNA_POWER_ON  do { P2OUT &= ~0x10; } while(0) // Enable power for LNA (P24, low-active)
#define LNA_POWER_OFF do { P2OUT |=  0x10; } while(0) // Disable power for LNA (P24, low-active)
#define ADR2INT(x)  (*((UINT16*)& x))

#define MAX_RXNUM_TABLE     (20)    // Length of rxNumTable.
#define MAX_RSSI_VALUES     (4)

typedef struct {                        
  UINT16 id;
  UINT8 num;
  UINT8 rxps[MAX_RSSI_VALUES];
  UINT8 rxpi;
  UINT16 rxp;
} rxNumTableEntry;

volatile packet_t rxPacket; // Public struct to a received packet.
volatile packet_t sentPacket;
rxNumTableEntry rxNumTable[MAX_RXNUM_TABLE];
UINT8 rxNumTablePos;        // Next pos to enter a new value into ::rxNumTable.

union {
 struct { UINT8 b2; UINT8 b1; UINT8 b4; UINT8 b3; };
 struct { UINT16 i1; UINT16 i2; };
} shiftbuf;

#define PRINT_STACK do {              \
   UINT16* p;                         \
   for(p= (UINT16*)READ_SP; p<=(UINT16*)0x38FF; p++){   \
     Comm_print("%x: %x\n", p, *p);               \
   }                                  \
   Comm_print("=================");    \
 } while(0);  

// DEBUG VERSIONS
#define RESTART_TX_TIMER()  do {    \
    if(GET_ISR(TIMERB0_VECTOR)==TX_STOPPED || GET_ISR(TIMERB0_VECTOR)==TX_DONE){ \
      Comm_print("##RETXTI %x %x##\n", GET_ISR(TIMERB0_VECTOR), READ_PC);        \
      PRINT_STACK;                                                    \
      System_reset();                                                        \
    }                                                                 \
    TBCCR0 = TBR + RADIO_STROKE;  \
    TBCCTL0 &= ~CCIFG;            \
    TBCCTL0 |= CCIE;              \
    } while(0)

#define STOP_TX_TIMER()  do { \
    if(GET_ISR(TIMERB0_VECTOR)==TX_WAITING){ \
      Comm_print("##STOXTI %x %x##\n", GET_ISR(TIMERB0_VECTOR), READ_PC);        \
      PRINT_STACK;                                                    \
      System_reset();                                                        \
    }                                                                 \
    TBCCTL0 &= ~(CCIFG+CCIE); \
  }while(0)


interrupt(NOVECTOR) TX_STOPPED() {
  System_stopWatchdog();
  PRINT_STACK;          // ADD
  System_reset();       // ADD
}

interrupt(NOVECTOR) TX_IDLE() {
  UINT16 i;
  UINT8 x;
  // If IDLE and something is in buf prepare packet in txBuf and
  // go to txState=TX_WAITING to wait for a free channel.
  if(numPackets > 0){
    // print("(%i bytes in txp)", RB_LENGTH(txPackets));
    // Prepare txBuf with packet to send:
    // 8 byte preamble and sync word
    // 8 byte header (in ringbuf txPackets)
    // x byte data   ( "    "       "     )
    // 2 byte checksm( "    "       "     )
    // dont forget to take 2 bytes callback off ringbuf
    // Start with 6 bytes preamble + 2 bytes sync word.
    txBuf[0] = 0xAA;
    txBuf[1] = 0xAA;
    txBuf[2] = 0xAA;
    txBuf[3] = 0xAA;
    txBuf[4] = 0xAA;
    txBuf[5] = 0xAA;
    txBuf[6] = 0xD3;
    txBuf[7] = 0x91;
    // Fetch header
    for(i=0; i<8; i++) {
      RB_GET(txPackets, x);
      txBuf[8+i] = x;
    }
    // Write header to sent packet
    sentPacket.to     = ADR2INT(txBuf[8+2]);
    sentPacket.from   = ADR2INT(txBuf[8+4]);
    sentPacket.type   = txBuf[8+6];
    sentPacket.num    = txBuf[8+7];
    sentPacket.header_length = 0;
    sentPacket.header = 0;
    sentPacket.data_length = txBuf[8+1];
    sentPacket.data   = NULL;
    // Now queue rest of data (length is contained in header).
    for(i=0; i<txBuf[9]+2; i++) { // ATTENTION CONSTANT 9! +2 for checksum
      RB_GET(txPackets, x);
      txBuf[16+i] = x;
    }
    // Add 2 Bytes FF as puffer. (?checksum?)
    txBuf[16+txBuf[9]+2] = 0xFA;       //txBuf[i++] = 0xFE;
    txBuf[16+txBuf[9]+3] = 0xFA;       //txBuf[i++] = 0xFE;
    // calc txbufLength: L + Preamble+Header + Cs + Puffer ?!
    txBufLength = txBuf[9] + 16 + 2 + 2           +2; //last +2??
    RB_GET(txPackets, x); txCallback  = (x << 8);       //tmp << 8;
    RB_GET(txPackets, x); txCallback |= x;  //tmp;// drop callback
    numPackets--;
    timesRepeated = 0;
    // go into waiting
    if(txMode & TXMODE_BACKOFF) {
      SET_ISR(TIMERB0_VECTOR, TX_WAITING);
    }
    else { // txMode==TXMODE_NOBACKOFF
      SET_ISR(TIMERB0_VECTOR, TX_BACKOFF);
      txBackoff = 1; // No backoff
      // Dont clear if here so state TX_BACKOFF will run
      // directly after this interrupt
    }
    // Go directly to TX_WAITING or TX_BACKOFF
    TBCCTL0 &=  ~(CCIE + CCIFG);
    TBCCTL0 |= CCIE + CCIFG;
  } 
  else {
    // If no packet has to be sent stop here and do nothing.
    // TimerTx interrupt will be reenabled if a packet is queued by
    // radio_queue_packet.
    STOP_TX_TIMER();
    SET_ISR(TIMERB0_VECTOR, TX_STOPPED);
    //RadioOFF
    if(txMode & TXMODE_LP) {
      CC1020_setupPD();
      LNA_POWER_OFF;
    }
  }
}

// Channel busy, waiting for free channel.
// case TX_WAITING: look for free chanel (rxState = RX_SEARCHING) and go to TX_BACKOFF
interrupt(NOVECTOR) TX_WAITING() {
  if(GET_ISR(USART0RX_VECTOR)==rxsearching && (CC_LOCK)) {
    txBackoff = (((TBR + TAR) & 0x001E) + 4);
    SET_ISR(TIMERB0_VECTOR, TX_BACKOFF); 
  }
  RESTART_TX_TIMER();
}

// Channel clear, waiting for random backoff time elapsed.
// case TX_BACKOFF: wait until backoff time elapsed, then start sending
interrupt(NOVECTOR) TX_BACKOFF() {
  // If we are currently sending an ack (no matter how this
  // could happen?!)  go back to TX_WAITING.
  if(GET_ISR(DACDMA_VECTOR)==acktxdma) {
        SET_ISR(TIMERB0_VECTOR, TX_WAITING);  
        RESTART_TX_TIMER();
  }
  else if((GET_ISR(USART0RX_VECTOR)!=rxsearching || (!(CC_LOCK))) && (txMode & TXMODE_BACKOFF)) {
        SET_ISR(TIMERB0_VECTOR, TX_WAITING);
        RESTART_TX_TIMER();
  }
  else if(txBackoff==0) { // Start sending.
    // REENTRANT!, case needs longer than 7ms
    STOP_TX_TIMER();
    DISABLE_RX_IRQ(); 
    _NOP();
    eint();  
    v0 = TAR;
    isSending = 1;
    // Switch transceiver to Tx
    LNA_POWER_OFF;
    CC1020_wakeupTX(TX_CURRENT);
    txLockStatus = CC1020_setupTX(TX_CURRENT);
    // Configure ports for Tx, Assumed UART is in SPI mode for chipcon receive
    P3SEL &= ~0x02;  // Ensure Rx line is off
    P3SEL |=  0x0C;  // select Tx line and clk
    U0CTL |= SWRST;  // NEW
    IFG1 &= ~UTXIFG0; // Reset IFG.
    DMACTL0 = DMA0TSEL_4; // Configure DMA Channel 0 for UART0 TXIFG.
    DMACTL1 = 0x0000; // No DMAONFETCH, ROUNDROBIN, ENNMI.
    // Single transfer Mode, dstadr unchanged, srcadr incremented, byte access
    // Important to use DMALEVEL when using USART TX interrupts so first edge 
    // doesn't get lost (hangs every 50. - 100. time)!
    DMA0CTL = DMADT_0 + DMADSTINCR_0 + DMASRCINCR_3 + DMASBDB + DMALEVEL + DMAIE;
    DMA0SA = (UINT16) txBuf;
    DMA0DA = (UINT16) &TXBUF0;
    DMA0SZ = txBufLength;
    U0CTL &= ~SWRST;  // NEW, needed??
    DMA0CTL |= DMAEN; // Start tx, when finished ends in radiotxdma
  }
  // Backoff time still running, decrease counter and restart timer.
  else {
      txBackoff--;       
      RESTART_TX_TIMER();
  }
}

interrupt(NOVECTOR) radiotxdma() {
  DMA0CTL &= ~(DMAIFG + DMAIE);
  eint(); // eint part lasts ~7ms
  /////////////////// Reconfigure Port& for receive mode.
  P3SEL &= ~0x04;  // Ensure Tx line is off
  P3SEL |=  0x0A;  // select rx line and clk
  /////////////////// Switch transceiver to Rx
  CC1020_wakeupRX(RX_CURRENT);
  CC1020_setupRX(RX_CURRENT) ;
  LNA_POWER_ON;
  dint();  
  if(*((UINT16*)&txBuf[10]) != BROADCAST) { // Not-Broadcasts need ack
    ackTimeout = 12; //ACK_TIMEOUT_19; // currentCC1020Config.ackTimeout; // ATTENTION!!
    SET_ISR(TIMERB0_VECTOR, TX_WAITACK);
    RESTART_TX_TIMER();
  } 
  else { // Broadcasts dont need a ack, signal "sent"
    SET_ISR(TIMERB0_VECTOR, TX_DONE);
    txBuf[0] = 0x00; // mark for no packet in txbuf 
    txCallbackValue = 1;                        
    WAKEUPLPM1(MF_RADIO_TX);
  }
  isSending = 0; // no more sending
  ENABLE_RX_IRQ();
}


interrupt(NOVECTOR) TX_DONE() {
  // nothing happens here, just to mark this state
  System_stopWatchdog();       
  PRINT_STACK;          
  System_reset();              
}

// Currently waiting for an ACK.
// case TX_WAITACK: repeat packet if needed, go to idle if ack received.
interrupt(NOVECTOR) TX_WAITACK() {
  ackTimeout--;
  if(ackTimeout==0) {
    if(timesRepeated > MAX_REPEATED) { // Unreachable destination
      SET_ISR(TIMERB0_VECTOR, TX_DONE);
      txBuf[0] = 0x00; // mark for no packet in txbuf
      txCallbackValue = 0;                        
      WAKEUPLPM1(MF_RADIO_TX);
    }
    else {
      timesRepeated++;
      SET_ISR(TIMERB0_VECTOR, TX_WAITING); 
      RESTART_TX_TIMER();
    }
  }
  else RESTART_TX_TIMER();
}

// Receiving side: Low level
//
// Preamble is 0xAA, Sync is 0xD391,
// Packet format:  AAAAAAAAAA D391 00 Length[1] To[2] From[2] Type[1] Num[1] 
//  
// 0  AA D3 91 00 | FF 00 |
// 1  55 69 C8 80 | 7F 80 |
// 2  AA B4 E4 40 | 3F C0 |
// 3  55 5A 72 20 | 1F E0 |
// 4  AA AD 39 10 | 0F F0 |
// 5  55 56 9C 88 | 07 F8 |
// 6  AA AB 4E 44 | 03 FC |
// 7  55 55 A7 22 | 01 FE |
// 8  AA AA D3 91 | 00 FF |
//
interrupt(NOVECTOR) acktxdma() {
  DMA0CTL &= ~(DMAIFG + DMAIE);
  eint(); // eint part lasts ~7ms
  /////////////////// Reconfigure Port& for receive mode.
  P3SEL &= ~0x04;  // Ensure Tx line is off
  P3SEL |=  0x0A;  // select rx line and clk
  /////////////////// Switch transceiver to Rx
  CC1020_wakeupRX(RX_CURRENT);
  CC1020_setupRX(RX_CURRENT);
  LNA_POWER_ON;
  dint();  
  _NOP();
  isSending = 0; // no more sending
  SET_ISR(DACDMA_VECTOR, radiotxdma);    
  ENABLE_RX_IRQ();
}

interrupt(NOVECTOR) rxsearching() {
  shiftbuf.b1 = shiftbuf.b2;
  shiftbuf.b2 = shiftbuf.b3;
  shiftbuf.b3 = shiftbuf.b4;
  shiftbuf.b4 = RXBUF0;
  if(shiftbuf.i1==0xAAD3 && shiftbuf.i2==0x9100) { // 0  AA D3 91 00 | FF 00 |
    SET_ISR(USART0RX_VECTOR, rxsynched0); 
    rxBuf[rxBufPos++] = shiftbuf.b4;
  }
  else if(shiftbuf.i1==0x5569 && shiftbuf.i2==0xC880) { // 1  55 69 C8 80 | 7F 80 |
    SET_ISR(USART0RX_VECTOR, rxsynched1); 
  }
  else if(shiftbuf.i1==0xAAB4 && shiftbuf.i2==0xE440) { // 2  AA B4 E4 40 | 3F C0 |
    SET_ISR(USART0RX_VECTOR, rxsynched2); 
  }
  else if(shiftbuf.i1==0x555A && shiftbuf.i2==0x7220) { // 3  55 5A 72 20 | 1F E0 |
    SET_ISR(USART0RX_VECTOR, rxsynched3);       
  }
  else if(shiftbuf.i1==0xAAAD && shiftbuf.i2==0x3910) { // 4  AA AD 39 10 | 0F F0 |
    SET_ISR(USART0RX_VECTOR, rxsynched4);         
  }
  else if(shiftbuf.i1==0x5556 && shiftbuf.i2==0x9C88) { // 5  55 56 9C 88 | 07 F8 |
    SET_ISR(USART0RX_VECTOR, rxsynched5);        
  }
  else if(shiftbuf.i1==0xAAAB && shiftbuf.i2==0x4E44) { // 6  AA AB 4E 44 | 03 FC |
    SET_ISR(USART0RX_VECTOR, rxsynched6); 
  }
  else if(shiftbuf.i1==0x5555 && shiftbuf.i2==0xA722) { // 7  55 55 A7 22 | 01 FE |
    SET_ISR(USART0RX_VECTOR, rxsynched7);      
  }
  // Signal "Channel busy"
  if(GET_ISR(USART0RX_VECTOR) != rxsearching){
    rxTime = CURRENT_TIME;
    rssi = CC1020_readRegister(CC1020_RSS);
  } 
}

interrupt(NOVECTOR) rxsynched0() { // L
  rxBuf[rxBufPos++] = RXBUF0;
  if(rxBufPos >= 8 && rxBufPos >= rxBuf[1]+8+4) SET_ISR(USART0RX_VECTOR, rxreceived); 
}
       
interrupt(NOVECTOR) rxsynched1() { // L <<1
  shiftbuf.b3 = shiftbuf.b4;
  shiftbuf.b4 = RXBUF0;
  shiftbuf.i1 = shiftbuf.i2 << 1;
  rxBuf[rxBufPos++] = shiftbuf.b1;
  if(rxBufPos >= 8 && rxBufPos >= rxBuf[1]+8+4) SET_ISR(USART0RX_VECTOR, rxreceived); 
}
       
interrupt(NOVECTOR) rxsynched2() { // L <<2
  shiftbuf.b3 = shiftbuf.b4;
  shiftbuf.b4 = RXBUF0;
  shiftbuf.i1 = shiftbuf.i2 << 2;
  rxBuf[rxBufPos++] = shiftbuf.b1;
  if(rxBufPos >= 8 && rxBufPos >= rxBuf[1]+8+4) SET_ISR(USART0RX_VECTOR, rxreceived); 
}
       
interrupt(NOVECTOR) rxsynched3() { // L <<3
  shiftbuf.b3 = shiftbuf.b4;
  shiftbuf.b4 = RXBUF0;
  shiftbuf.i1 = shiftbuf.i2 << 3;
  rxBuf[rxBufPos++] = shiftbuf.b1;
  if(rxBufPos >= 8 && rxBufPos >= rxBuf[1]+8+4) SET_ISR(USART0RX_VECTOR, rxreceived); 
}
       
interrupt(NOVECTOR) rxsynched4() { // L <<4
  shiftbuf.b3 = shiftbuf.b4;
  shiftbuf.b4 = RXBUF0;
  shiftbuf.i1 = shiftbuf.i2 << 4;
  rxBuf[rxBufPos++] = shiftbuf.b1;
  if(rxBufPos >= 8 && rxBufPos >= rxBuf[1]+8+4) SET_ISR(USART0RX_VECTOR, rxreceived); 
}
       
interrupt(NOVECTOR) rxsynched5() { // R >>3
  shiftbuf.b3 = shiftbuf.b4;
  shiftbuf.b4 = RXBUF0;
  shiftbuf.i1 = shiftbuf.i2 >> 3;
  rxBuf[rxBufPos++] = shiftbuf.b2;       
  if(rxBufPos >= 8 && rxBufPos >= rxBuf[1]+8+4) SET_ISR(USART0RX_VECTOR, rxreceived); 
}
       
interrupt(NOVECTOR) rxsynched6() { // R >>2
  shiftbuf.b3 = shiftbuf.b4;
  shiftbuf.b4 = RXBUF0;
  shiftbuf.i1 = shiftbuf.i2 >> 2;
  rxBuf[rxBufPos++] = shiftbuf.b2;       
  if(rxBufPos >= 8 && rxBufPos >= rxBuf[1]+8+4) SET_ISR(USART0RX_VECTOR, rxreceived); 
}
       
interrupt(NOVECTOR) rxsynched7() { // R >>1
  shiftbuf.b3 = shiftbuf.b4;
  shiftbuf.b4 = RXBUF0;
  shiftbuf.i1 = shiftbuf.i2 >> 1;
  rxBuf[rxBufPos++] = shiftbuf.b2;       
  if(rxBufPos >= 8 && rxBufPos >= rxBuf[1]+8+4) SET_ISR(USART0RX_VECTOR, rxreceived); 
}

void _Net_sendACK(UINT8* buf) {
  // Sending an ACK, first thing is to check if DMA controller
  // is already in use, if it is the TX state machine i currently sending
  // a packet. Should not happen when a recent packet was received, but
  // for safety:
  if(DMA0CTL & DMAEN) return;
  // Next thing is to set the ISR to acktxdma, this will block 
  // the tx machine from starting another dma transfer.
  SET_ISR(DACDMA_VECTOR, acktxdma); 
  // Start with 6 bytes preamble + 2 bytes sync word.
  ackBuf[0]  = 0xAA;
  ackBuf[1]  = 0xAA;
  ackBuf[2]  = 0xAA;
  ackBuf[3]  = 0xAA;
  ackBuf[4]  = 0xAA;
  ackBuf[5]  = 0xAA;
  ackBuf[6]  = 0xD3;
  ackBuf[7]  = 0x91;
  // Queue header
  ackBuf[8]               = 0x00;
  ackBuf[9]               = 0x00;                  // Length
  *((UINT16*)&ackBuf[10]) = *((UINT16*)&buf[4]);   // To
  *((UINT16*)&ackBuf[12]) = Configuration.id;      // From
  ackBuf[14] = ACK_PACKET;                         // Type
  ackBuf[15] = buf[7];                             // Num
  ackBuf[16] = 0xFF;
  ackBuf[17] = 0xFF;
  isSending = 1;
  DISABLE_RX_IRQ(); 
  _NOP();
  eint();
  // Switch transceiver to Tx
  LNA_POWER_OFF;
  CC1020_wakeupTX(TX_CURRENT);
  txLockStatus = CC1020_setupTX(TX_CURRENT) ;
  // Configure ports for Tx, Assumed UART is in SPI mode for chipcon receive
  P3SEL &= ~0x02; // Ensure Rx line is off
  P3SEL |=  0x0C; // select Tx line and clk
  U0CTL |= SWRST; // NEW    TODO Uart reinit?
  IFG1 &= ~UTXIFG0;     // Reset IFG.
  DMACTL0 = DMA0TSEL_4; // Configure DMA Channel 0 for UART0 TXIFG.
  DMACTL1 = 0x0000;     // No DMAONFETCH, ROUNDROBIN, ENNMI.
  // Single transfer Mode, dstadr unchanged, srcadr incremented, byte access
  // Important to use DMALEVEL when using USART TX interrupts so first edge 
  // doesn't get lost (hangs every 50. - 100. time)!
  DMA0CTL = DMADT_0 + DMADSTINCR_0 + DMASRCINCR_3 + DMASBDB + DMALEVEL + DMAIE;
  DMA0SA = (UINT16) ackBuf;
  DMA0DA = (UINT16) &TXBUF0;
  DMA0SZ = 18;
  // TODO ????
  // dint(); _NOP();
  U0CTL &= ~SWRST; // NEW
  DMA0CTL |= DMAEN;
}

interrupt(NOVECTOR) rxreceived() {
  UINT16 i;
  // calculate Checksum
  checksumR = ((rxBuf[rxBuf[1]+9])<<8) + (rxBuf[rxBuf[1]+8]);
  checksumC = INIT_FCS;
  for(i=1; i< 8 + rxBuf[1]; i++) checksumC = _Net_ccittFcs(checksumC, rxBuf[i]);
  if(rxBuf[6]==ACK_PACKET) { // Received an ack
    // Waiting for this ack?!? This part perhaps not irq safe ?! TODO
    if((GET_ISR(TIMERB0_VECTOR) == TX_WAITACK) && (*((UINT16*)&rxBuf[4]) == *((UINT16*)&txBuf[10])) && (rxBuf[7]==txBuf[15])) {
      STOP_TX_TIMER();
      SET_ISR(TIMERB0_VECTOR, TX_DONE);
      txBuf[0] = 0x00; // mark for no packet in txbuf
      txCallbackValue = 1;                        
      WAKEUPLPM1(MF_RADIO_TX);
    }
  }
  else if(((*((UINT16*)&rxBuf[2])) == Configuration.id) && checksumR == checksumC) {
    _Net_sendACK(rxBuf);
  }        
  dint(); _NOP();
  rxBuf[0] = 0x01;
  if(rxBuf==rxBuf1) rxBuf=rxBuf2;
  // TODO check and warn if buffer is not already empty
  else rxBuf=rxBuf1;
  rxBufPos = 0;
  SET_ISR(USART0RX_VECTOR, rxsearching); 
  WAKEUPLPM1(MF_RADIO_RX);
  IE1 |= URXIE0;        // Enable interrupt.
  IFG1 &= ~URXIFG0;
  DIRQEND();
}

void _Net_packetHandler(UINT8* fullBuf) {
  UINT8 i;
  char c;
  // Fill struct
  rxPacket.to     = ADR2INT(fullBuf[2]);
  rxPacket.from   = ADR2INT(fullBuf[4]);
  rxPacket.type   = fullBuf[6];
  rxPacket.num    = fullBuf[7];
  rxPacket.header_length = 0;
  rxPacket.header = 0;
  rxPacket.data_length = fullBuf[1];
  rxPacket.data   = (UINT8*) &fullBuf[8];
  // Debug output
  if(Comm_logLevel & DEBUG_RADIO) {
    if(rxPacket.type==ACK_PACKET) c = 'A';
    else if(checksumR==checksumC) c = '+';
    else                          c = '-';
    Comm_print("[%c T:%x F:%x T:%b N:%i L:%i R:%i Cr:%x Cc:%x\r\n", 
           c,
           rxPacket.to, rxPacket.from, rxPacket.type, 
           rxPacket.num, rxPacket.data_length, rssi, checksumR, checksumC); 
  }
  if(checksumR==checksumC) {
    // Packet already received?
    for(i=0; i<MAX_RXNUM_TABLE; i++) {
      if(rxNumTable[i].id == rxPacket.from && rxNumTable[i].num == rxPacket.num) return; // We already received that, so do nothing.
    }
    // First time received: enter current num into history.
    for(i=0; i<=MAX_RXNUM_TABLE; i++) {
      if(i==MAX_RXNUM_TABLE) { // If no entry found from rxPacket.from create a new one..
        if(rxNumTablePos==MAX_RXNUM_TABLE) return;
        rxNumTable[rxNumTablePos].id = rxPacket.from;
        rxNumTable[rxNumTablePos].num = rxPacket.num; 
        rxNumTable[rxNumTablePos].rxps[0] = rssi;
        rxNumTable[rxNumTablePos].rxps[1] = rssi;
        rxNumTable[rxNumTablePos].rxps[2] = rssi;
        rxNumTable[rxNumTablePos].rxps[3] = rssi;
        rxNumTable[rxNumTablePos].rxpi = 0;
        rxNumTable[rxNumTablePos].rxp = rssi << 2;
        rxNumTablePos++;
        if(rxNumTablePos==MAX_RXNUM_TABLE) rxNumTablePos = 0;           
      }
      else if(rxNumTable[i].id == rxPacket.from) { // else just overwrite old number.
        rxNumTable[i].num = rxPacket.num;
        rxNumTable[i].rxp -= rxNumTable[i].rxps[ rxNumTable[i].rxpi ];
        rxNumTable[i].rxp += rssi;
        rxNumTable[i].rxps[ rxNumTable[i].rxpi ] = rssi;
        rxNumTable[i].rxpi++;
        if(rxNumTable[i].rxpi>=MAX_RSSI_VALUES) rxNumTable[i].rxpi = 0;
        break;
      }
    }
	if(rxPacket.type == ACK_PACKET) return;
	if(callbacks[C_RADIO]) callbacks[C_RADIO](NULL);
  }
}

// name UART0 Switching functions
void _Net_resetU0() {
  //Reset SEL for P3[1-3] (CC DIO, DIO, DCLK) and P3[4-5] (Camera Rx+Tx)
  P3SEL &= ~0x3E;             
  IFG1 &= ~(UTXIE0 + URXIE0);
  ME1 &= ~(UTXE0 + URXE0);    // Disable Uart0 Tx + Rx
  UCTL0  = SWRST;             // U0 into reset state.
}

/*
 *   brief Initializes UART0 for receiving from Chipcon
 *
 *   Data (for transmit+receive) must be stable at the rising edge of the clock.
 *   This means:
 *     Transmit : Set bit on falling edge of clock.
 *     Receive  : Read bit on rising edge of clock.
 *
 *   CKPH: Normal UCLK clocking scheme.
 *   CKPL: 1 The inactive level is high; data is output with the falling edge of
 *         UCLK; input data is latched with the rising edge of UCLK.
 */
void _Net_initU0() {
  _Net_resetU0();
  UCTL0 |= CHAR + SYNC; // 8-bit character, SPI, Slave mode
  UTCTL0 = CKPL +  STC; // CKPH works also, but not CKPH+CKPL or none of them!!
  URCTL0 = 0x00;
  UBR00 = 0x00;         // No baudrate divider 
  UBR10 = 0x00;         // settings for a spi
  UMCTL0 = 0x00;        // slave.
  ME1 |= USPIE0;        // Enable USART0 TXD/RXD, disabling does not yield any powersavings
  P3SEL |=  0x0A;       // Select rx line and clk
  UCTL0  &= ~SWRST;     // Last thing is clear reset bit
  IE1 |= URXIE0;        // Enable interrupt.
}

bool Net_rxHandler() {
  //////////////////// Look into both buffers for something to handle
  while(rxBuf1[0]==0x01 || rxBuf2[0]==0x01) {
    if(rxBuf1[0]==0x01) {
      _Net_packetHandler(rxBuf1);
      rxBuf1[0] = 0xFF; // Set buffer empty again.
    } 
    if(rxBuf2[0]==0x01) {
      _Net_packetHandler(rxBuf2);
      rxBuf2[0] = 0xFF;
    }
  }
  runModule &= ~MF_RADIO_RX;
  return True;
}

bool Net_txHandler() {
  if(txCallback != 0) ((fp_char_t)txCallback)(txCallbackValue);
  SET_ISR(TIMERB0_VECTOR, TX_IDLE);       
  runModule &= ~MF_RADIO_TX;
  RESTART_TX_TIMER();
  return True;
}

bool Net_send(packet_t* packet, fp_char_t callback) {
  UINT16 i;
  UINT16 cb = (UINT16) callback;
  UINT16 length = packet->header_length + packet->data_length;
  UINT16 checksum = INIT_FCS;
  if(!radioEnabled) return False;
  // todo IF not enabled RETURN!!
  if(length & 0x0001) { //if length is uneven make it even
    length++;
    packet->data_length++;
  }
  if(254 - RB_LENGTH(txPackets) > (length + 8 + 2 + 2)) { // we have space in buf, so queue it, length + 8 data bytes, 2 checksum and 2 callback
    // inc txnum
    txNum++;
    // Queue header
    RB_PUT(txPackets, 0x00); // to l
    //checksum = ccitt_fcs(checksum, 0x00);
    RB_PUT(txPackets, length); // to l
    checksum = _Net_ccittFcs(checksum, length);
    RB_PUT(txPackets, packet->to); // to l
    checksum = _Net_ccittFcs(checksum,packet->to);
    RB_PUT(txPackets, packet->to >> 8); // to h      
    checksum = _Net_ccittFcs(checksum, packet->to >> 8);
    RB_PUT(txPackets, packet->from); // from l
	checksum = _Net_ccittFcs(checksum, packet->from);
    RB_PUT(txPackets, packet->from >> 8); // from h
    checksum = _Net_ccittFcs(checksum, packet->from >> 8);
    RB_PUT(txPackets, packet->type); // type
    checksum = _Net_ccittFcs(checksum, packet->type);
    RB_PUT(txPackets, txNum); // num
    checksum = _Net_ccittFcs(checksum, txNum);
    // Queue data header and data
    for(i=0; i<packet->header_length; i++) { 
      RB_PUT(txPackets, packet->header[i]); 
      checksum = _Net_ccittFcs(checksum, packet->header[i]);
    }
    for(i=0; i<packet->data_length; i++) {   
      RB_PUT(txPackets, packet->data[i]);
      checksum = _Net_ccittFcs(checksum, packet->data[i]);
    }
    RB_PUT(txPackets, checksum); 
    RB_PUT(txPackets, checksum >> 8); 
    // Queue function callback 
    RB_PUT(txPackets, cb >> 8); // upper 8 bits of callback
    RB_PUT(txPackets, cb);      // lower 8 bits of callback  
    numPackets++;
    // Debug output
    if(Comm_logLevel & DEBUG_RADIO) {
      Comm_print("[Q T:%x F:%x T:%b N:%i L:%i Cc:%x]\r\n", 
             packet->to, packet->from, packet->type, 
             txNum, packet->data_length, checksum);           
    }    
    // If first packet in buf immediatly start Interrupt for sending
    if(GET_ISR(TIMERB0_VECTOR)==TX_STOPPED) {
      //RadioON
      if(txMode & TXMODE_LP) {
        UINT16 t1, t2, t3, t4;
        t1 = TAR;
        t3 = TBR;
        CC1020_wakeupRX(RX_CURRENT);
        CC1020_calibrate();     
        CC1020_wakeupTX(TX_CURRENT);
        CC1020_calibrate();
        CC1020_wakeupRX(RX_CURRENT);
        CC1020_setupRX(RX_CURRENT) ;
        LNA_POWER_ON;
        t2 = TAR;
        t4 = TBR;
      }
      dint();
	  _NOP();
      TBCCTL0 &= ~(CCIE + CCIFG);
      SET_ISR(TIMERB0_VECTOR, TX_IDLE);  
      TBCCTL0 |= CCIE + CCIFG; 
    }
    eint();
    return True;        
  }
  else { // Not enough space in ringbuf!
    if(Comm_logLevel & DEBUG_RADIO) 
      Comm_print("[N T:%x F:%x T:%b N:%i L:%i Cc:%x\r\n", 
             packet->to, packet->from, packet->type, 
             packet->num, packet->data_length, checksum);      
       /// TODO  2005-12-7 ONLY QUICK HACK TO REBOOT DEVICE WHEN TX STATE MACHINE IS HANGING IN TX_WAITING:
       if(GET_ISR(TIMERB0_VECTOR) == TX_WAITING) {
         Comm_print("##TX_MACHINE HANGING IN TX_WAITING!! RESET\n\n\n");
         System_reset();
       }
       return False;  
  }
}

UINT8 Net_getRSSI(UINT16 node) {
  UINT16 i;
  for(i=0; i<rxNumTablePos; i++) {
    if(rxNumTable[i].id==node) return (rxNumTable[i].rxp>>2);  
  }
  return 0;
}

bool Net_init(){
  UINT16 i;
  radioEnabled = 0;
  irqLock=0;
  isSending = 0;
  rxNumTablePos = 0;
  for(i=0; i<MAX_RXNUM_TABLE; i++) {
    rxNumTable[i].id   = 0;
    rxNumTable[i].num  = 0;
    rxNumTable[i].rxps[0] = 0;
    rxNumTable[i].rxps[1] = 0;
    rxNumTable[i].rxps[2] = 0;
    rxNumTable[i].rxpi = 0;
    rxNumTable[i].rxp = 0;
  }
  CC1020_init();
  return True;
}

bool Net_on(UINT8 mode) {
  if(radioEnabled) return True;
  LNA_POWER_ON;
  // Transmitting side vars
  txMode = mode;
  numPackets = 0;
  txBackoff = 1; //Init
  // Receiving side vars
  rxBuf1[0] = 0xFF;
  rxBuf2[0] = 0xFF;
  rxBuf = rxBuf1;
  rxBufPos = 0;
  // Interrupt routings
  SET_ISR(USART0RX_VECTOR, rxsearching); 
  STOP_TX_TIMER(); 
  SET_ISR(TIMERB0_VECTOR,  TX_STOPPED);
  SET_ISR(DACDMA_VECTOR, radiotxdma); 
  // Ring buffers  
  RB_INIT(txPackets);
  // Uart
  _Net_initU0();
  CC1020_setupRX(RX_CURRENT);  
  if(txMode & TXMODE_LP) {
    CC1020_setupPD();
    LNA_POWER_OFF;      
  }
  radioEnabled=1;
  return True;
}

bool Net_changeMode(UINT8 mode) {
  txMode = mode;
  if(!(mode & TXMODE_LP)) {
    CC1020_wakeupRX(RX_CURRENT);
    CC1020_calibrate();     
    CC1020_wakeupTX(TX_CURRENT);
    CC1020_calibrate();
    CC1020_wakeupRX(RX_CURRENT);
    CC1020_setupRX(RX_CURRENT) ;
    LNA_POWER_ON;  
  }
  return True;
}

bool Net_off(){
   if(!radioEnabled) return True;
   dint();
   _NOP(); 
   TBCCTL0 &= ~(CCIFG+CCIE);
   IE1 &= ~(URXIE0); 
   SET_ISR(TIMERB0_VECTOR,  TX_STOPPED);
   eint();
   CC1020_setupPD();
   LNA_POWER_OFF;
   radioEnabled = 0;
   return True;
}

bool Net_setTxPower(UINT8 n) {
  Configuration.transmitPower = n;
  CC1020_paPower = n;
  return True;
}
