/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

/** \file ScatterWeb.Time.h
 ** \ingroup System
 ** \brief The ScatterWeb.Time represents an instant in time, typically
 ** expressed as a date and time of day.
 ** <p>
 ** This file keeps the global time of this node in the global var currentTime.
 ** It contains methods for writing, reading and converting currentTime.
 ** To avoid concurrency problems accessing currentTime must be done with
 ** disabled interrupts or in interrupt routines only. This defines some macro
 ** for accessing Time_currentTime safely.  
 **
 ** @author        Marco Hutta
 ** @version       1.0
 ** @since         1.0
 **/ 

#ifndef ScatterWeb_Time
#define ScatterWeb_Time

/**
 * Reads current time with disabled interrupts.
 *
 * @since         1.0
 */
#define CURRENT_TIME \
({ \
  extern volatile time_t currentTime; \
  time_t time;  \
  dint(); \
  _NOP(); \
  time = currentTime; \
  eint(); \
  time; \
}) 

/**
 * Reads the UINT32 secs from current time with disabled interrupts.
 *
 * @since         1.0
 */
#define CURRENT_TIME_SECS \
({ \
  extern volatile time_t currentTime; \
  UINT32 secs;  \
  dint(); \
  _NOP(); \
  secs = currentTime.secs; \
  eint(); \
  secs; \
}) 

/**
 * struct of time
 * <ul>
 * <li>year
 * <li>month
 * <li>day
 * <li>hour
 * <li>min
 * <li>sec
 * </ul>
 *  
 * @since         1.0
 */
typedef struct {
  UINT8 year;
  UINT8 month;
  UINT8 day;
  UINT8 hour;
  UINT8 min;
  UINT8 sec;
} btime_t;

/**
 * struct of system time
 * <ul>
 * <li>secs
 * <li>millis
 * </ul>
 *  
 * @since         1.0
 */
typedef struct {
  UINT32 secs;
  UINT16 millis;
} time_t;

/**
 * Read currentTime or safely with disabled interrupts. Easier (and also faster)
 * to use is CURRENT_TIME: 
 * time_t time = CURRENT_TIME; <-> time_t time; getCurrentTime(&time); 
 *
 * @param time      current time struct to store system time
 * @return          pointer of time
 * @since           1.0
 */
time_t* Time_get(time_t* time);

/**
 * Write currentTime or safely with disabled interrupts.
 *
 * @param time      current time to set 
 * @return          <code>True</code> always.
 * @since           1.0
 */
bool Time_set(time_t* time);

/**
 * Converts broken down time to time
 *
 * @param src       input time
 * @param dst       output current time 
 * @return          <code>True</code> always.
 * @since           1.0
 */
bool Time_convertTime(const time_t* src, btime_t* dst);

/**
 * Converts time to broken down time
 *
 * @param src       input current time
 * @param dst       output btime
 * @return          <code>True</code> always.
 * @since           2.0
 */
bool Time_convertBtime(const btime_t* src, time_t* dst);

/**
 * Writes broken down time as a string in the form yyyy-mm-dd hh:mm:ss
 *
 * @param writer    function to write
 * @param time      time to write
 * @return          <code>True</code> always.
 * @since           1.0
 */
bool Time_write(fp_char_t writer, btime_t* time);

/**
 * Reads broken down time as a string.
 * The format is the same as in Time_writeTime, but the year must be written
 * without 2000. The number of digits is not fixed like in Time_writeTime, so it
 * is ok to write 03-09-07 08:31:02 or 3-9-7 8:31:2.
 *
 * @param c         array to store time
 * @param time      time to read
 * @return          <code>True</code> always.
 * @since           1.0
 */
bool Time_read(const UINT8* c, btime_t* time);

#endif
